#!/projects/QM/local/bin/python3

'''
Created on Sep 5, 2016

@author: ayakovenko
'''

import sys
import numpy as np


#This function upload  voxel maps as binary file (for y_eye)
def import_voxel_map(input_voxel_map_file_name) :
    Y_MAGIC = np.int32(2004)
    size_m = np.int32(0); size_x = np.int32(0); size_y = np.int32(0); size_z = np.int32(0); vmap = np.object()
    center_x = np.float32(0.); center_y = np.float32(0.); center_z = np.float32(0.); spacering = np.float32(0.)
    if input_voxel_map_file_name is None :
        print("ERROR. The given input voxel map file name is empty!")
        return (False, size_m, size_x, size_y, size_z, center_x, center_y, center_z, spacering, vmap) 
    try:
        file_in = open(input_voxel_map_file_name, mode = "rb", buffering = 2048)
    except IOError:
        print ("ERROR. Could not read file {:s}".format(input_voxel_map_file_name))
        return (False, size_m, size_x, size_y, size_z, center_x, center_y, center_z, spacering, vmap) 
    try: #Read binary voxel map file
        bytestring = file_in.read(np.dtype(np.int32).itemsize)
        if bytestring == '' or np.int32(np.fromstring(bytestring, dtype=np.int32, count=1, sep='')[0]) != Y_MAGIC : print("ERROR during reading int32 Y_MAGIC.")
        bytestring = file_in.read(np.dtype(np.float32).itemsize);
        if bytestring != '' : center_x = np.float32(np.fromstring(bytestring, dtype=np.float32, count=1, sep='')[0])
        else : print("ERROR during reading float32 center_x.")
        bytestring = file_in.read(np.dtype(np.float32).itemsize);
        if bytestring != '' : center_y = np.float32(np.fromstring(bytestring, dtype=np.float32, count=1, sep='')[0])
        else : print("ERROR during reading float32 center_y.")
        bytestring = file_in.read(np.dtype(np.float32).itemsize);
        if bytestring != '' : center_z = np.float32(np.fromstring(bytestring, dtype=np.float32, count=1, sep='')[0])
        else : print("ERROR during reading float32 center_z.")
        bytestring = file_in.read(np.dtype(np.int32).itemsize);
        if bytestring != '' : size_m = np.int32(np.fromstring(bytestring, dtype=np.int32, count=1, sep='')[0])
        else : print("ERROR during reading int32 size_m.")
        bytestring = file_in.read(np.dtype(np.int32).itemsize);
        if bytestring != '' : size_x = np.int32(np.fromstring(bytestring, dtype=np.int32, count=1, sep='')[0])
        else : print("ERROR during reading int32 size_x.")
        bytestring = file_in.read(np.dtype(np.int32).itemsize);
        if bytestring != '' : size_y = np.int32(np.fromstring(bytestring, dtype=np.int32, count=1, sep='')[0])
        else : print("ERROR during reading int32 size_y.")
        bytestring = file_in.read(np.dtype(np.int32).itemsize);
        if bytestring != '' : size_z = np.int32(np.fromstring(bytestring, dtype=np.int32, count=1, sep='')[0])
        else : print("ERROR during reading int32 size_z.")
        bytestring = file_in.read(np.dtype(np.float32).itemsize);
        if bytestring != '' : spacering = np.float32(np.fromstring(bytestring, dtype=np.float32, count=1, sep='')[0])
        else : print("ERROR during reading float32 spacering.")
        bytestring = file_in.read(np.dtype(np.int8).itemsize*size_m*size_x*size_y*size_z);
        if bytestring != '' : vmap = np.fromstring(bytestring, dtype=np.int8, count=size_m*size_x*size_y*size_z, sep='')
        else : print("ERROR during reading {:d}x{:d}x{:d}x{:d} vmap.".format(size_m, size_x, size_y, size_z))
    except IOError:
        print ("ERROR. Could not write to file {:s}".format(input_voxel_map_file_name))
        return (False, size_m, size_x, size_y, size_z, center_x, center_y, center_z, spacering, vmap)
    finally: 
        file_in.close()
    return (True, size_m, size_x, size_y, size_z, center_x, center_y, center_z, spacering, vmap) 

def show_help_message(name, input0_voxel_map_file_name, input1_voxel_map_file_name, verbose, flag) :
    print("{:s} - a python script for creating a voxel maps of  molecular complexes from corresponding pdbqt files.\n" \
          "The usage is:\n" \
          "\t -i0 {:s} -- (lexem) input voxel map file name #0.\n" \
          "\t -i1 {:s} -- (lexem) input voxel map file name #1.\n" \
          "\t -v {:s} -- (verbose) be loud and noisy.\n"
          "\t -h {:s} -- show this help and exit.\n".format(name, input0_voxel_map_file_name, input1_voxel_map_file_name, str(verbose), str(flag)))    

#The main function of py_box
if __name__ == '__main__':
    
    #Stage I. Interface
    input0_voxel_map_file_name = "input0.vmap"
    input1_voxel_map_file_name = "input1.vmap"
    count = np.int32(1); flag = np.bool8(False); verbose = np.bool8(False)
    while count != len(sys.argv) :
        if   len(sys.argv[count]) == len("-h")  and sys.argv[count] == "-h"  : flag = True
        elif len(sys.argv[count]) == len("-i0") and sys.argv[count] == "-i0" :
            count += 1
            if count != len(sys.argv) : input0_voxel_map_file_name = sys.argv[count] 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a filename.\n".format(count)); flag = True; break
        elif len(sys.argv[count]) == len("-i1") and sys.argv[count] == "-i1" : 
            count += 1
            if count != len(sys.argv) : input1_voxel_map_file_name = sys.argv[count] 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a filename.\n".format(count)); flag = True; break
        elif len(sys.argv[count]) == len("-v")  and sys.argv[count] == "-v"  : verbose = True
        else : print("ERROR. Can't recognize {:d} input line lexem.\n".format(count)); flag = True; break
        count += 1
    if flag or verbose : 
        show_help_message(sys.argv[0], input0_voxel_map_file_name, input1_voxel_map_file_name, verbose, flag)
        if flag : sys.exit(2)

    #Stage II. Upload vmaps
    (flag, size_m0, size_x0, size_y0, size_z0, center_x0, center_y0, center_z0, spacering0, vmap0) = import_voxel_map(input0_voxel_map_file_name)
    if not flag : sys.exit("ERROR. Can't upload the first voxel map from file {:s}.".format(input0_voxel_map_file_name))
    if verbose : print("INFO. The first  voxel map {:d}x{:d}x{:d}x{:d} centered at {{{:f},{:f},{:f}}} with spacering {:f} was successfuly uploaded from file {:s}.".format(size_m0, size_x0, size_y0, size_z0, center_x0, center_y0, center_z0, spacering0, input0_voxel_map_file_name))
    (flag, size_m1, size_x1, size_y1, size_z1, center_x1, center_y1, center_z1, spacering1, vmap1) = import_voxel_map(input1_voxel_map_file_name)
    if not flag : sys.exit("ERROR. Can't upload the second voxel map from file {:s}.".format(input1_voxel_map_file_name))
    if verbose : print("INFO. The second voxel map {:d}x{:d}x{:d}x{:d} centered at {{{:f},{:f},{:f}}} with spacering {:f} was successfuly uploaded from file {:s}.".format(size_m1, size_x1, size_y1, size_z1, center_x1, center_y1, center_z1, spacering1, input1_voxel_map_file_name))
    if size_m0 != size_m1 or size_x0 != size_x1 or size_y0 != size_y1 or size_z0 != size_z1 or center_x0 != center_x1 or center_y0 != center_y1 or center_z0 != center_z1 or spacering0 != spacering1 : sys.exit("ERROR. The imported vmaps are NOT compatible!")

    #Stage III. Compare vmaps
    m = np.int32(); i = np.int32(); j = np.int32(); k = np.int32()
    rmsd = np.float64(0.); l = np.uint32(0)
    for m in range(0, size_m0) :
        for i in range(0, size_x0) :
            for j in range(0, size_y0) :
                for k in range(0, size_z0) :
                    rmsd += (np.float64(np.int32(vmap0[l]) - np.int32(vmap1[l])) * (np.int32(vmap0[l]) - np.int32(vmap1[l]))) / 10000.; l += 1

    #Stage IV. Show info and exit
    print("{:f} is the {:d}x{:d}x{:d}x{:d} vmaps RMSD between {:s} and {:s}.".format(rmsd, size_m0, size_x0, size_y0, size_z0, input0_voxel_map_file_name, input1_voxel_map_file_name))
    sys.exit(0)


