#!/projects/QM/local/bin/python3

'''
Created on Sep 1, 2016

@author: ayakovenko
'''

import sys
import numpy as np
import math


#This function defines internal types of gro atoms
def get_gro_chem_type(ch_type) :
    i = np.int32(0) 
    while i != 4 and ch_type[i] == ' ' : i += 1 
    if i != 4 : 
        if   ch_type[i] == 'H' : return (np.int8(1), np.float32(0.60), np.float32(1.09) ) # H, 0.25A actually
        elif ch_type[i] == 'C' :
            if i != 3 and (ch_type[i + 1] == 'l' or ch_type[i + 1] == 'L') : return ( np.int8(2), np.float32(1.00), np.float32(1.75) ) # Cl 
            else               : return ( np.int8(2), np.float32(1.00), np.float32(1.70) ) # C, rather CH3 group
        elif ch_type[i] == 'N' : return ( np.int8(3), np.float32(0.65), np.float32(1.56) ) # N
        elif ch_type[i] == 'O' : return ( np.int8(4), np.float32(0.60), np.float32(1.52) ) # O
        elif ch_type[i] == 'F' : return ( np.int8(4), np.float32(0.50), np.float32(1.47) ) # F 
        elif ch_type[i] == 'P' : return ( np.int8(5), np.float32(1.00), np.float32(1.80) ) # P
        elif ch_type[i] == 'S' : return ( np.int8(5), np.float32(1.00), np.float32(1.80) ) # S
        elif ch_type[i] == 'B' :
            if i != 3 and (ch_type[i + 1] == 'r' or ch_type[i +1] == 'R') : return ( np.int8(2), np.float32(1.15), np.float32(1.85) ) # Br
        elif ch_type[i] == 'I' : return ( np.int8(2), np.float32(1.40), np.float32(1.98) ) # I
    return (False, 0.0, 0.0)

#The function to define string object type
def isint(value) :
    try:
        x=int(value)
        return (True,  x)
    except ValueError :
        return (False, 0)

#The function to define string object type
def isfloat(value) :
    try:
        x=float(value)
        return (True,  x)
    except ValueError :
        return (False, float('nan'))

#This function removes nonpolar hydrogen considering distances to (bound) atoms
def remove_nonpolar_hydrogens(size_atoms, atoms) : 
    MAX_H_BOND_LENGTH = 0.115**2 #in nm!!!
    i = np.int32()
    j = np.int32()
    k = np.int32()
    for i in range(0, size_atoms) :
        if atoms[i]['ch_type'] == 1 :
            flag = False
            if i > size_atoms - i : #As hydrogen's neighbour is likely to be nearby, screen to left and right simulaneously
                j = 1
                while (j < (size_atoms - i)) :
                    k = i + j #Forward step
                    if (atoms[k]['ch_type']) and (atoms[k]['ch_type'] != 1) and (((atoms[i]['x'] - atoms[k]['x'])**2 + (atoms[i]['y'] - atoms[k]['y'])**2 + (atoms[i]['z'] - atoms[k]['z'])**2) <= MAX_H_BOND_LENGTH) :
                        if ( (atoms[k]['ch_type'] != 3) and (atoms[k]['ch_type'] != 4) and (atoms[k]['ch_type'] != 5) ) :
                            atoms[i]['ch_type'] = 0
                        flag=True; break;
                    k = i - j #Backward step, skip (eliminated) hydrogens
                    if (atoms[k]['ch_type']) and (atoms[k]['ch_type'] != 1) and (((atoms[i]['x'] - atoms[k]['x'])**2 + (atoms[i]['y'] - atoms[k]['y'])**2 + (atoms[i]['z'] - atoms[k]['z'])**2) <= MAX_H_BOND_LENGTH) :
                        if ( (atoms[k]['ch_type'] != 3) and (atoms[k]['ch_type'] != 4) and (atoms[k]['ch_type'] != 5) ) :
                            atoms[i]['ch_type'] = 0
                        flag=True; break;
                    j += 1
                if not flag : 
                    for k in range (0, i - j) : #Remains of Backward step
                        if (atoms[k]['ch_type']) and (atoms[k]['ch_type'] != 1) and (((atoms[i]['x'] - atoms[k]['x'])**2 + (atoms[i]['y'] - atoms[k]['y'])**2 + (atoms[i]['z'] - atoms[k]['z'])**2) <= MAX_H_BOND_LENGTH) :
                            if ( (atoms[k]['ch_type'] != 3) and (atoms[k]['ch_type'] != 4) and (atoms[k]['ch_type'] != 5) ) :
                                atoms[i]['ch_type'] = 0
                            flag=True; break;
            else                   : 
                j = 1
                while (j < i) :
                    k = i + j #Forward step
                    if (atoms[k]['ch_type']) and (atoms[k]['ch_type'] != 1) and (((atoms[i]['x'] - atoms[k]['x'])**2 + (atoms[i]['y'] - atoms[k]['y'])**2 + (atoms[i]['z'] - atoms[k]['z'])**2) <= MAX_H_BOND_LENGTH) :
                        if ( (atoms[k]['ch_type'] != 3) and (atoms[k]['ch_type'] != 4) and (atoms[k]['ch_type'] != 5) ) :
                            atoms[i]['ch_type'] = 0
                        flag=True; break;
                    k = i - j #Backward step, skip 'eliminated' hydrogens
                    if (atoms[k]['ch_type']) and (atoms[k]['ch_type'] != 1) and (((atoms[i]['x'] - atoms[k]['x'])**2 + (atoms[i]['y'] - atoms[k]['y'])**2 + (atoms[i]['z'] - atoms[k]['z'])**2) <= MAX_H_BOND_LENGTH) :
                        if ( (atoms[k]['ch_type'] != 3) and (atoms[k]['ch_type'] != 4) and (atoms[k]['ch_type'] != 5) ) :
                            atoms[i]['ch_type'] = 0
                        flag=True; break;
                    j += 1
                if not flag : 
                    for k in range (i + j, size_atoms) : #Remains of Forward step
                        if (atoms[k]['ch_type']) and (atoms[k]['ch_type'] != 1) and (((atoms[i]['x'] - atoms[k]['x'])**2 + (atoms[i]['y'] - atoms[k]['y'])**2 + (atoms[i]['z'] - atoms[k]['z'])**2) <= MAX_H_BOND_LENGTH) :
                            if ( (atoms[k]['ch_type'] != 3) and (atoms[k]['ch_type'] != 4) and (atoms[k]['ch_type'] != 5) ) :
                                atoms[i]['ch_type'] = 0
                            flag=True; break;
    j = 0 #Remove marked hydrogens
    for i in range(0, size_atoms) :
        if atoms[i]['ch_type'] :
            atoms[j]['ch_type'] = atoms[i]['ch_type']; atoms[j]['r_a'] = atoms[j]['r_a']; atoms[j]['r_vdw'] = atoms[i]['r_vdw']; atoms[j]['x'] = atoms[i]['x']; atoms[j]['y'] = atoms[i]['y']; atoms[j]['z'] = atoms[i]['z']
            j += 1
    return j
             
#This function loads molecule into a set of atoms with coordinates into a structure
#('ch_type', np.int8), ('r_vdw', np.float32), ('x', np.float32), ('y', np.float32), ('z', np.float32)]
def import_gro(file_name, verbose=True) :
    #Stage I. Import data from pdbqt file
    if file_name is None :
        print("ERROR. The given import gro file name is empty!")
        return (False, None, None)
    try:
        file_in = open(file_name, mode = "r", buffering = 2048)
    except IOError:
        print ("ERROR. Could not open file {:s} for reading".format(file_name))
        return (False, None, None)
    #Stage II. Initialize data storages
    size_ratoms  = np.int32(0)
    ratoms = np.empty(shape=(0,), dtype=[('ch_type', np.int8), ('r_a', np.float32), ('r_vdw', np.float32), ('x', np.float32), ('y', np.float32), ('z', np.float32)], order='C')
    size_latoms  = np.int32(0)
    latoms = np.empty(shape=(0,), dtype=[('ch_type', np.int8), ('r_a', np.float32), ('r_vdw', np.float32), ('x', np.float32), ('y', np.float32), ('z', np.float32)], order='C')
    line_len   = np.int32()
    count      = np.int32(0)
    flag       = np.bool8(False)
    #Stage III. Read the model
    try:
        for line in file_in:
            count += 1
            line_len = len(line)
            if (count == 1) or ( line_len <= 44 ) : continue; #Skip short lines and gro header
            if ((line[5]=='D') or (line[5]=='d')) and ((line[6]=='R') or (line[6]=='r')) and ((line[7] == 'G') or (line[7] == 'g')) :
                if not (size_latoms % 0xFF) : latoms.resize((size_latoms+0xFF,), refcheck=False)
                (flag, latoms[size_latoms]['x']) = isfloat(line[20:28])
                if not flag : 
                    if verbose : print("Warning. Can't parse atoms' x-coordinate in line {:d} of file {:s}. The line is skipped.".format(count, file_name)); continue
                (flag, latoms[size_latoms]['y']) = isfloat(line[28:36])
                if not flag : 
                    if verbose : print("Warning. Can't parse atoms' y-coordinate in line {:d} of file {:s}. The line is skipped.".format(count, file_name)); continue
                (flag, latoms[size_latoms]['z']) = isfloat(line[36:44])
                if not flag :
                    if verbose : print("Warning. Can't parse atoms' z-coordinate in line {:d} of file {:s}. The line is skipped.".format(count, file_name)); continue
                (latoms[size_latoms]['ch_type'], latoms[size_latoms]['r_a'], latoms[size_latoms]['r_vdw']) = get_gro_chem_type(line[11:15])
                if not latoms[size_latoms]['ch_type'] : 
                    if verbose : print("Warning. Can't parse atoms' chemical type in line {:d} of file {:s}. The line is skipped.".format(count, file_name)); continue
                size_latoms += 1
            else : #Read receptor atom
                if not (size_ratoms % 0xFF) : ratoms.resize((size_ratoms+0xFF,), refcheck=False)
                (flag, ratoms[size_ratoms]['x']) = isfloat(line[20:28])
                if not flag : 
                    if verbose : print("Warning. Can't parse atoms' x-coordinate in line {:d} of file {:s}. The line is skipped.".format(count, file_name)); continue
                (flag, ratoms[size_ratoms]['y']) = isfloat(line[28:36])
                if not flag : 
                    if verbose : print("Warning. Can't parse atoms' y-coordinate in line {:d} of file {:s}. The line is skipped.".format(count, file_name)); continue
                (flag, ratoms[size_ratoms]['z']) = isfloat(line[36:44])
                if not flag :
                    if verbose : print("Warning. Can't parse atoms' z-coordinate in line {:d} of file {:s}. The line is skipped.".format(count, file_name)); continue
                (ratoms[size_ratoms]['ch_type'], ratoms[size_ratoms]['r_a'], ratoms[size_ratoms]['r_vdw']) = get_gro_chem_type(line[11:15])
                if not ratoms[size_ratoms]['ch_type'] : 
                    if verbose : print("Warning. Can't parse atoms' chemical type in line {:d} of file {:s}. The line is skipped.".format(count, file_name)); continue
                size_ratoms += 1
    except IOError:
        print ("ERROR. Could not read file {:s}".format(file_name))
        del line; del line_len; del count; del flag; del latoms; del size_latoms; del ratoms; del size_ratoms   
        file_in.close(); del file_in
        return (False, None)
    #Free memory 
    del line; del line_len; del count; del flag
    file_in.close(); del file_in
    #Remove nonpolar hydrogens
    count = size_latoms; size_latoms=remove_nonpolar_hydrogens(size_atoms=size_latoms, atoms=latoms) 
    if verbose : print("INFO. {:d} nonpolar hydrogens (out of {:d} all atoms) were removed from ligand   structure.\n".format(count - size_latoms, count))
    for count in range(0, size_latoms) : latoms[count]['ch_type'] -= 1; latoms[count]['x'] *= 10.; latoms[count]['y'] *= 10.; latoms[count]['z'] *= 10.
    count = size_ratoms; size_ratoms=remove_nonpolar_hydrogens(size_atoms=size_ratoms, atoms=ratoms)
    if verbose : print("INFO. {:d} nonpolar hydrogens (out of {:d} all atoms) were removed from receptor structure.\n".format(count - size_ratoms, count))
    for count in range(0, size_ratoms) : ratoms[count]['ch_type'] -= 1; ratoms[count]['x'] *= 10.; ratoms[count]['y'] *= 10.; ratoms[count]['z'] *= 10.
    latoms.resize((size_latoms,), refcheck=False); del size_latoms
    ratoms.resize((size_ratoms,), refcheck=False); del size_ratoms
    return (True, latoms, ratoms)
       
#This function fill spatial voxel map 
#It encodes floats in range -1., ..., 0, ..., +1. as chars in range -100, ..., 0, ... +100
def fill_voxel_map(sign, spacering, center_x, center_y, center_z, size_m, size_x, size_y, size_z, vmap, atoms) :
    if sign != +1 and sign != -1 : print("ERROR. Given sign attribute is neither +1 nor -1."); return False
    count = np.int32(0); flag = np.bool8(); r_vdw2 = np.float32(); r2 = np.float32(); N_DECR = np.int32(); N_INCR = np.int32()
    p = np.empty(shape=(3,), dtype=[('x', np.int32),   ('y', np.int32),   ('z', np.int32)  ], order='C')  
    d = np.empty(shape=(3,), dtype=[('x', np.float32), ('y', np.float32), ('z', np.float32)], order='C')
    #Here we assume size_x, size_y and size_z are even so the center is at the mesh intersection line
    if not (size_x % 2) : d[2]['x'] = spacering * (size_x / 2); d[0]['x'] = center_x - d[2]['x']; d[1]['x'] = center_x + d[2]['x'] 
    else :  d[2]['x'] = spacering * (((size_x - 1) / 2) + 0.5); d[0]['x'] = center_x - d[2]['x']; d[1]['x'] = center_x + d[2]['x'] 
    if not (size_y % 2) : d[2]['y'] = spacering * (size_y / 2); d[0]['y'] = center_y - d[2]['y']; d[1]['y'] = center_y + d[2]['y'] 
    else :  d[2]['y'] = spacering * (((size_y - 1) / 2) + 0.5); d[0]['y'] = center_y - d[2]['y']; d[1]['y'] = center_y + d[2]['y'] 
    if not (size_z % 2) : d[2]['z'] = spacering * (size_z / 2); d[0]['z'] = center_z - d[2]['z']; d[1]['z'] = center_z + d[2]['z'] 
    else :  d[2]['z'] = spacering * (((size_z - 1) / 2) + 0.5); d[0]['z'] = center_z - d[2]['z']; d[1]['z'] = center_z + d[2]['z'] 
    if sign == +1 :
        for n in range(0, atoms.shape[0]) :
            if ( (atoms[n]['x'] > d[0]['x']) and (atoms[n]['x'] < d[1]['x']) and
                 (atoms[n]['y'] > d[0]['y']) and (atoms[n]['y'] < d[1]['y']) and
                 (atoms[n]['z'] > d[0]['z']) and (atoms[n]['z'] < d[1]['z']) ) :
                N_DECR = -(+atoms[n]['r_vdw'] // spacering); N_INCR = 2 - N_DECR #N_INCR is counting for range(beg, end-1) bellow
                d[2]['x'] = (atoms[n]['x'] - d[0]['x']); d[2]['y'] = (atoms[n]['y'] - d[0]['y']); d[2]['z'] = (atoms[n]['z'] - d[0]['z'])
                r_vdw2 = atoms[n]['r_vdw']**2; r_a2 = atoms[n]['r_a']**2; NORM_S = (atoms[n]['r_vdw'] - atoms[n]['r_a']) / 2.; NORM_EXPK = -1. / 2. / NORM_S / NORM_S
                p[2]['x'] = d[2]['x'] // spacering; p[0]['x'] = p[2]['x'] + N_DECR; p[1]['x'] = p[2]['x'] + N_INCR 
                if p[0]['x'] < 0      : p[0]['x'] = 0 
                if p[1]['x'] > size_x : p[1]['x'] = size_x
                p[2]['y'] = d[2]['y'] // spacering; p[0]['y'] = p[2]['y'] + N_DECR; p[1]['y'] = p[2]['y'] + N_INCR
                if p[0]['y'] < 0      : p[0]['y'] = 0
                if p[1]['y'] > size_y : p[1]['y'] = size_y
                p[2]['z'] = d[2]['z'] // spacering; p[0]['z'] = p[2]['z'] + N_DECR; p[1]['z'] = p[2]['z'] + N_INCR
                if p[0]['z'] < 0      : p[0]['z'] = 0
                if p[1]['z'] > size_z : p[1]['z'] = size_z
                flag = False
                for i in range (p[0]['x'], p[1]['x']) :
                    for j in range (p[0]['y'], p[1]['y']) :
                        for k in range (p[0]['z'], p[1]['z']) :
                            r2 = (np.float32(i) * spacering - d[2]['x'])**2 + (np.float32(j) * spacering - d[2]['y'])**2 + (np.float32(k) * spacering - d[2]['z'])**2
                            if r2 < r_vdw2 :
                                flag = True  
                                if r2 > r_a2 :
                                    c = np.int8(100. * math.exp(NORM_EXPK * (math.sqrt(r2) - atoms[n]['r_a'])**2))    
                                    if c > abs(vmap[((atoms[n]['ch_type'] * size_x + i) * size_y + j) * size_z + k]) :
                                        vmap[((atoms[n]['ch_type'] * size_x + i) * size_y + j) * size_z + k] = c                              
                                else : vmap[((atoms[n]['ch_type'] * size_x + i) * size_y + j) * size_z + k] = np.int8(+100)
                if flag : count += 1
    else          :
        for n in range(0, atoms.shape[0]) :
            if ( (atoms[n]['x'] > d[0]['x']) and (atoms[n]['x'] < d[1]['x']) and
                 (atoms[n]['y'] > d[0]['y']) and (atoms[n]['y'] < d[1]['y']) and
                 (atoms[n]['z'] > d[0]['z']) and (atoms[n]['z'] < d[1]['z']) ) :
                N_DECR = -(+atoms[n]['r_vdw'] // spacering); N_INCR = 2 - N_DECR #N_INCR is counting for range(beg, end-1) bellow
                d[2]['x'] = (atoms[n]['x'] - d[0]['x']); d[2]['y'] = (atoms[n]['y'] - d[0]['y']); d[2]['z'] = (atoms[n]['z'] - d[0]['z'])
                r_vdw2 = atoms[n]['r_vdw']**2; r_a2 = atoms[n]['r_a']**2; NORM_S = (atoms[n]['r_vdw'] - atoms[n]['r_a']) / 2.; NORM_EXPK = -1. / 2. / NORM_S / NORM_S
                p[2]['x'] = d[2]['x'] // spacering; p[0]['x'] = p[2]['x'] + N_DECR; p[1]['x'] = p[2]['x'] + N_INCR 
                if p[0]['x'] < 0      : p[0]['x'] = 0 
                if p[1]['x'] > size_x : p[1]['x'] = size_x
                p[2]['y'] = d[2]['y'] // spacering; p[0]['y'] = p[2]['y'] + N_DECR; p[1]['y'] = p[2]['y'] + N_INCR
                if p[0]['y'] < 0      : p[0]['y'] = 0
                if p[1]['y'] > size_y : p[1]['y'] = size_y
                p[2]['z'] = d[2]['z'] // spacering; p[0]['z'] = p[2]['z'] + N_DECR; p[1]['z'] = p[2]['z'] + N_INCR
                if p[0]['z'] < 0      : p[0]['z'] = 0
                if p[1]['z'] > size_z : p[1]['z'] = size_z
                flag = False 
                for i in range (p[0]['x'], p[1]['x']) :
                    for j in range (p[0]['y'], p[1]['y']) :
                        for k in range (p[0]['z'], p[1]['z']) :
                            r2 = (np.float32(i) * spacering - d[2]['x'])**2 + (np.float32(j) * spacering - d[2]['y'])**2 + (np.float32(k) * spacering - d[2]['z'])**2
                            if r2 < r_vdw2 :
                                flag = True  
                                if r2 > r_a2 :
                                    c = np.int8(100. * math.exp(NORM_EXPK * (math.sqrt(r2) - atoms[n]['r_a'])**2))    
                                    if c > abs(vmap[((atoms[n]['ch_type'] * size_x + i) * size_y + j) * size_z + k]) :
                                        vmap[((atoms[n]['ch_type'] * size_x + i) * size_y + j) * size_z + k] = -c                              
                                else : vmap[((atoms[n]['ch_type'] * size_x + i) * size_y + j) * size_z + k] = np.int8(-100)
                if flag : count += 1
    return count


#This function writes voxel maps as binary file (for y_peye)
def write_voxel_maps_bin(voxel_map_file_name, spacering, center_x, center_y, center_z, size_m, size_x, size_y, size_z, vmap) :
    if voxel_map_file_name is None :
        print("ERROR. The given export voxel map file name is empty!")
        return False
    try:
        file_out = open(voxel_map_file_name, mode = "wb", buffering = 2048)
    except IOError:
        print ("ERROR. Could not read file {:s}".format(voxel_map_file_name))
        return False
    Y_MAGIC = np.int32(2004)
    try: #Write voxels to pdb
        if file_out.write(Y_MAGIC.tobytes()) != np.dtype(np.int32).itemsize : print("ERROR during conversion/writing int32."); return False
        if file_out.write(center_x.tobytes()) != np.dtype(np.float32).itemsize : print("ERROR during conversion/writing float32."); return False
        if file_out.write(center_y.tobytes()) != np.dtype(np.float32).itemsize : print("ERROR during conversion/writing float32."); return False
        if file_out.write(center_z.tobytes()) != np.dtype(np.float32).itemsize : print("ERROR during conversion/writing float32."); return False
        if file_out.write(size_m.tobytes()) != np.dtype(np.int32).itemsize : print("ERROR during conversion/writing int32."); return False
        if file_out.write(size_x.tobytes()) != np.dtype(np.int32).itemsize : print("ERROR during conversion/writing int32."); return False
        if file_out.write(size_y.tobytes()) != np.dtype(np.int32).itemsize : print("ERROR during conversion/writing of int32."); return False
        if file_out.write(size_z.tobytes()) != np.dtype(np.int32).itemsize : print("ERROR during conversion/writing of int32."); return False
        if file_out.write(spacering.tobytes()) != np.dtype(np.float32).itemsize : print("ERROR during conversion/writing float32."); return False
        if file_out.write(vmap.tobytes())   != np.dtype(np.int8).itemsize * size_m * size_x * size_y * size_z : print("Conversion/writing error of vmap array."); return False
    except IOError:
        print ("ERROR. Could not write to file {:s}".format(voxel_map_file_name))
        return False
    finally: 
        file_out.close()
    return True

#This function writes voxel maps in human-friendly format
def write_voxel_maps_pdb(voxel_pdb_file_name, spacering, center_x, center_y, center_z, size_m, size_x, size_y, size_z, vmap) :
    if voxel_pdb_file_name is None :
        print("ERROR. The given export voxel pdb file name is empty!")
        return False
    try:
        file_out = open(voxel_pdb_file_name, mode = "w", buffering = 2048)
    except IOError:
        print ("ERROR. Could not read file {:s}".format(voxel_pdb_file_name))
        return False
    count = np.int32(0); i = np.int32(); j = np.int32(); k = np.int32()
    try: #Write voxels to pdb
        #Receptor map
        for i in range (0, size_x) :
            for j in range (0, size_y) :
                for k in range (0, size_z) :
                    #Save hydrogen map  
                    if vmap[((0 * size_x + i) * size_y + j) * size_z + k] >= np.int8(+25) :
                        count += 1; file_out.write("ATOM  {:5d}  H       H   1    {:8.3f}{:8.3f}{:8.3f}\n".format(count,center_x - spacering * (size_x / 2 - i), center_y - spacering * (size_y / 2 - j), center_z - spacering * (size_z / 2 - k))) 
        for i in range (0, size_x) :
            for j in range (0, size_y) :
                for k in range (0, size_z) :
                    #Save hydrogen map  
                    if vmap[((0 * size_x + i) * size_y + j) * size_z + k] <= np.int8(-25) :
                        count += 1; file_out.write("ATOM  {:5d}  H       H   2    {:8.3f}{:8.3f}{:8.3f}\n".format(count,center_x - spacering * (size_x / 2 - i), center_y - spacering * (size_y / 2 - j), center_z - spacering * (size_z / 2 - k))) 
        for i in range (0, size_x) :
            for j in range (0, size_y) :
                for k in range (0, size_z) :
                    #Save carbon map  
                    if vmap[((1 * size_x + i) * size_y + j) * size_z + k] >= np.int8(+25) :
                        count += 1; file_out.write("ATOM  {:5d}  C       C   1    {:8.3f}{:8.3f}{:8.3f}\n".format(count,center_x - spacering * (size_x / 2 - i), center_y - spacering * (size_y / 2 - j), center_z - spacering * (size_z / 2 - k))) 
        for i in range (0, size_x) :
            for j in range (0, size_y) :
                for k in range (0, size_z) :
                    #Save carbon map  
                    if vmap[((1 * size_x + i) * size_y + j) * size_z + k] <= np.int8(-25) :
                        count += 1; file_out.write("ATOM  {:5d}  C       C   2    {:8.3f}{:8.3f}{:8.3f}\n".format(count,center_x - spacering * (size_x / 2 - i), center_y - spacering * (size_y / 2 - j), center_z - spacering * (size_z / 2 - k))) 
        for i in range (0, size_x) :
            for j in range (0, size_y) :
                for k in range (0, size_z) :
                    #Save nitrogen map  
                    if vmap[((2 * size_x + i) * size_y + j) * size_z + k] >= np.int8(+25) :
                        count += 1; file_out.write("ATOM  {:5d}  N       N   1    {:8.3f}{:8.3f}{:8.3f}\n".format(count,center_x - spacering * (size_x / 2 - i), center_y - spacering * (size_y / 2 - j), center_z - spacering * (size_z / 2 - k))) 
        for i in range (0, size_x) :
            for j in range (0, size_y) :
                for k in range (0, size_z) :
                    #Save nitrogen map  
                    if vmap[((2 * size_x + i) * size_y + j) * size_z + k] <= np.int8(-25) :
                        count += 1; file_out.write("ATOM  {:5d}  N       N   2    {:8.3f}{:8.3f}{:8.3f}\n".format(count,center_x - spacering * (size_x / 2 - i), center_y - spacering * (size_y / 2 - j), center_z - spacering * (size_z / 2 - k))) 
        for i in range (0, size_x) :
            for j in range (0, size_y) :
                for k in range (0, size_z) :
                    #Save oxygen map  
                    if vmap[((3 * size_x + i) * size_y + j) * size_z + k] >= np.int8(+25) :
                        count += 1; file_out.write("ATOM  {:5d}  O       O   1    {:8.3f}{:8.3f}{:8.3f}\n".format(count,center_x - spacering * (size_x / 2 - i), center_y - spacering * (size_y / 2 - j), center_z - spacering * (size_z / 2 - k))) 
        for i in range (0, size_x) :
            for j in range (0, size_y) :
                for k in range (0, size_z) :
                    #Save oxygen map  
                    if vmap[((3 * size_x + i) * size_y + j) * size_z + k] <= np.int8(-25) :
                        count += 1; file_out.write("ATOM  {:5d}  O       O   2    {:8.3f}{:8.3f}{:8.3f}\n".format(count,center_x - spacering * (size_x / 2 - i), center_y - spacering * (size_y / 2 - j), center_z - spacering * (size_z / 2 - k))) 
        for i in range (0, size_x) :
            for j in range (0, size_y) :
                for k in range (0, size_z) :
                    #Save sulfur map  
                    if vmap[((4 * size_x + i) * size_y + j) * size_z + k] >= np.int8(+25) :
                        count += 1; file_out.write("ATOM  {:5d}  S       S   1    {:8.3f}{:8.3f}{:8.3f}\n".format(count,center_x - spacering * (size_x / 2 - i), center_y - spacering * (size_y / 2 - j), center_z - spacering * (size_z / 2 - k))) 
        for i in range (0, size_x) :
            for j in range (0, size_y) :
                for k in range (0, size_z) :
                    #Save sulfur map  
                    if vmap[((4 * size_x + i) * size_y + j) * size_z + k] <= np.int8(-25) :
                        count += 1; file_out.write("ATOM  {:5d}  S       S   2    {:8.3f}{:8.3f}{:8.3f}\n".format(count,center_x - spacering * (size_x / 2 - i), center_y - spacering * (size_y / 2 - j), center_z - spacering * (size_z / 2 - k))) 
    except IOError:
        print ("ERROR. Could not write to file {:s}".format(voxel_pdb_file_name))
        return False
    finally: 
        file_out.close()
    return True



def show_help_message(name, input_complex_file_name, output_voxel_map_file_name, output_voxel_pdb_file_name, center_x, center_y, center_z, size_x, size_y, size_z, spacering, verbose, flag) :
    if output_voxel_pdb_file_name is None : output_voxel_pdb_file_name = "None"
    print("{:s} - a python script for creating a voxel maps of  molecular complexes from corresponding gro files.\n" \
          "The usage is:\n" \
          "\t -if {:s} -- (lexem) input complex file name\n" \
          "\t -vf {:s} -- (lexem) output voxel map file name\n" \
          "\t -pf {:s} -- (lexem, optional) output pdb file of voxel map\n" \
          "\t -cx {:f} -- (real) x coordinate of box center\n" \
          "\t -cy {:f} -- (real) y coordinate of box center\n" \
          "\t -cz {:f} -- (real) z coordinate of box center\n" \
          "\t -sx {:d} -- (3*X+6 int) size of box in x dimension\n" \
          "\t -sy {:d} -- (3*Y+6 int) size of box in y dimension\n" \
          "\t -sz {:d} -- (3*Z+6 int) size of box in z dimension\n" \
          "\t -sp {:f} -- (positive real) box spatial resolution\n" \
          "\t -v {:s} -- (verbose) be loud and noisy.\n"
          "\t -h {:s} -- show this help and exit.\n".format(name, input_complex_file_name, output_voxel_map_file_name, output_voxel_pdb_file_name, center_x, center_y, center_z, size_x, size_y, size_z, spacering, str(verbose), str(flag)))    

#The main function of py_box
if __name__ == '__main__':
    
    #Stage I. Interface
    input_complex_file_name    = "complex.gro"
    output_voxel_map_file_name = "complex.vmap"
    output_voxel_pdb_file_name = None
    size_m = np.uint32(0); spacering = np.float32(1.)
    size_x = np.uint32(21); size_y = np.uint32(27); size_z = np.uint32(21)
    center_x = np.float32(-2.); center_y = np.float32(-2.); center_z = np.float32(+8.)
    count = np.int32(1); flag = np.bool8(); hflag = np.bool8(False); verbose = np.bool8(False)
    while count != len(sys.argv) :
        if   len(sys.argv[count]) == len("-h")  and sys.argv[count] == "-h"  : hflag = True
        elif len(sys.argv[count]) == len("-if") and sys.argv[count] == "-if" :
            count += 1
            if count != len(sys.argv) : input_complex_file_name = sys.argv[count] 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a filename.\n".format(count)); hflag = True; break
        elif len(sys.argv[count]) == len("-vf") and sys.argv[count] == "-vf" : 
            count += 1
            if count != len(sys.argv) : output_voxel_map_file_name = sys.argv[count] 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a filename.\n".format(count)); hflag = True; break
        elif len(sys.argv[count]) == len("-pf") and sys.argv[count] == "-pf" : 
            count += 1
            if count != len(sys.argv) : output_voxel_pdb_file_name = sys.argv[count] 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a filename.\n".format(count)); hflag = True; break
        elif len(sys.argv[count]) == len("-cx") and sys.argv[count] == "-cx" : 
            count += 1
            if count != len(sys.argv) :
                (flag, center_x) = isfloat(sys.argv[count])
                if not flag : print("ERROR. Given box center_x parameter is NOT a real.\n"); hflag = True; break; 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a real.\n"); hflag = True; break
        elif len(sys.argv[count]) == len("-cy") and sys.argv[count] == "-cy" : 
            count += 1
            if count != len(sys.argv) :
                (flag, center_y) = isfloat(sys.argv[count])
                if not flag : print("ERROR. Given box center_y parameter is NOT a real.\n"); hflag = True; break; 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a real.\n"); hflag = True; break
        elif len(sys.argv[count]) == len("-cz") and sys.argv[count] == "-cz" : 
            count += 1
            if count != len(sys.argv) :
                (flag, center_z) = isfloat(sys.argv[count])
                if not flag : print("ERROR. Given box center_z parameter is NOT a real.\n"); hflag = True; break; 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a real.\n"); hflag = True; break
        elif len(sys.argv[count]) == len("-sx") and sys.argv[count] == "-sx" : 
            count += 1
            if count != len(sys.argv) :
                (flag, size_x) = isint(sys.argv[count])
                if ( (not flag) or (size_x < 9) or ((size_x - 6) % 3) ) : print("ERROR. Given box size_x parameter is NOT 3*X+6 integer >=9.\n"); flag=True; break; 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a even positive integer.\n"); hflag = True; break
        elif len(sys.argv[count]) == len("-sy") and sys.argv[count] == "-sy" : 
            count += 1
            if count != len(sys.argv) :
                (flag, size_y) = isint(sys.argv[count])
                if ( (not flag) or (size_y < 9) or ((size_y - 6) % 3) ) : print("ERROR. Given box size_y parameter is NOT 3*Y+6 integer >=9.\n"); flag=True; break; 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a even positive integer.\n"); hflag = True; break
        elif len(sys.argv[count]) == len("-sz") and sys.argv[count] == "-sz" : 
            count += 1
            if count != len(sys.argv) :
                (flag, size_z) = isint(sys.argv[count])
                if ( (not flag) or (size_z < 9) or ((size_z - 6) % 3) ) : print("ERROR. Given box size_z parameter is NOT 3*Z+6 integer >=9.\n"); flag=True; break; 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a even positive integer.\n"); hflag = True; break
        elif len(sys.argv[count]) == len("-sp") and sys.argv[count] == "-sp" : 
            count += 1
            if count != len(sys.argv) :
                (flag, spacering) = isfloat(sys.argv[count])
                if ( (not flag) or (spacering <= 0.) ) : print("ERROR. Given box spacering parameter is NOT a positive real.\n"); flag=True; break; 
            else : print("ERROR. Can't parse {:d} input line lexem, should be a positive real.\n"); hflag = True; break
        elif len(sys.argv[count]) == len("-v")  and sys.argv[count] == "-v"  : verbose = True
        else : print("ERROR. Can't recognize {:d} input line lexem.\n".format(count)); hflag = True; break
        count += 1
    if hflag or verbose : 
        show_help_message(sys.argv[0], input_complex_file_name, output_voxel_map_file_name, output_voxel_pdb_file_name, center_x, center_y, center_z, size_x, size_y, size_z, spacering, verbose, hflag)
        if hflag : sys.exit(2)

    #Stage II. Init all
    size_m = 5
    vmap = np.zeros(shape=(size_m * size_x * size_y * size_z, ), dtype=np.int8, order='C')
    if verbose : print("INFO. A voxel map [{:d}][{:d}][{:d}][{:d}] was initialized.".format(size_m,size_x,size_y,size_z))

    #Stage III. Upload molecules
    (flag, latoms, ratoms) = import_gro(file_name=input_complex_file_name, verbose=verbose)
    if ((not flag) or (not ratoms.shape[0])) : sys.exit("ERROR. During reading receptor file {:s}.".format(input_complex_file_name))
    else       :
        if verbose : print("INFO. {:d} receptor and {:d} ligand atom records were successfully uploaded from receptor file name {:s}".format(ratoms.shape[0], latoms.shape[0], input_complex_file_name))

    #Stage V. Fill in voxel maps
    if verbose : print("Projecting atoms onto the voxel map centered at {:f}, {:f}, {:f}.".format(center_x, center_y, center_z))
    count = fill_voxel_map(sign=np.int8(+1), spacering=np.float32(spacering), center_x=np.float32(center_x), center_y=np.float32(center_y), center_z=np.float32(center_z), size_m=np.int32(size_m), size_x=np.int32(size_x), size_y=np.int32(size_y), size_z=np.int32(size_z), vmap=vmap, atoms=ratoms)
    if not count : sys.exit("ERROR. No receptor atoms were projected onto the voxel map.")
    else         :
        if verbose : print("INFO. {:d} receptor atoms were projected onto the voxel map.".format(count))
    count = fill_voxel_map(sign=np.int8(-1), spacering=np.float32(spacering), center_x=np.float32(center_x), center_y=np.float32(center_y), center_z=np.float32(center_z), size_m=np.int32(size_m), size_x=np.int32(size_x), size_y=np.int32(size_y), size_z=np.int32(size_z), vmap=vmap, atoms=latoms)
    if not count : sys.exit("ERROR. No ligand atoms were projected onto the voxel map.")
    else         :
        if verbose : print("INFO. {:d} ligand atoms were projected onto the voxel map.".format(count))

    #Stage VI. Save binary voxel map
    flag = write_voxel_maps_bin(voxel_map_file_name=output_voxel_map_file_name, spacering=np.float32(spacering), center_x=np.float32(center_x), center_y=np.float32(center_y), center_z=np.float32(center_z), size_m=np.int32(size_m), size_x=np.int32(size_x), size_y=np.int32(size_y), size_z=np.int32(size_z), vmap=vmap)
    if not flag : print("ERROR. Can't save voxels as a 3D bitmap file {:s}.".format(output_voxel_map_file_name)); sys.exit(2)
    #Stage VI.2. Save human version of voxel maps (for debugging)
    if output_voxel_pdb_file_name is not None :
        write_voxel_maps_pdb(output_voxel_pdb_file_name, spacering=np.float32(spacering), center_x=np.float32(center_x), center_y=np.float32(center_y), center_z=np.float32(center_z), size_m=np.int32(size_m), size_x=np.int32(size_x), size_y=np.int32(size_y), size_z=np.int32(size_z), vmap=vmap)
    #Finalization
    del(latoms); del(ratoms); del(vmap)
    if verbose : print("py_box is successfully completed!\n")
    sys.exit(0)

